@@ -22,6 +22,8 @@ import docx, docx.table
 from docx import Document
 from docx.shared import RGBColor, Pt, Inches
 from docx.enum.text import WD_COLOR, WD_ALIGN_PARAGRAPH
+from docx.oxml import OxmlElement
+from docx.oxml.ns import qn
 
 from bs4 import BeautifulSoup
 
@@ -30,6 +32,9 @@ INDENT = 0.25
 LIST_INDENT = 0.5
 MAX_INDENT = 5.5 # To stop indents going off the page
 
+# Style to use with tables. By default no style is used.
+DEFAULT_TABLE_STYLE = None
+
 def get_filename_from_url(url):
     return os.path.basename(urlparse(url).path)
 
@@ -58,9 +63,70 @@ def fetch_image(url):
 def remove_last_occurence(ls, x):
     ls.pop(len(ls) - ls[::-1].index(x) - 1)
 
-def remove_whitespace(string):
+def remove_whitespace(string, leading=False, trailing=False):
+    """Remove white space from a string.
+
+    Args:
+        string(str): The string to remove white space from.
+        leading(bool, optional): Remove leading new lines when True.
+        trailing(bool, optional): Remove trailing new lines when False.
+
+    Returns:
+        str: The input string with new line characters removed and white space squashed.
+
+    Examples:
+
+        Single or multiple new line characters are replaced with space.
+
+            >>> remove_whitespace("abc\\ndef")
+            'abc def'
+            >>> remove_whitespace("abc\\n\\n\\ndef")
+            'abc def'
+
+        New line characters surrounded by white space are replaced with a single space.
+
+            >>> remove_whitespace("abc \\n \\n \\n def")
+            'abc def'
+            >>> remove_whitespace("abc  \\n  \\n  \\n  def")
+            'abc def'
+
+        Leading and trailing new lines are replaced with a single space.
+
+            >>> remove_whitespace("\\nabc")
+            ' abc'
+            >>> remove_whitespace("  \\n  abc")
+            ' abc'
+            >>> remove_whitespace("abc\\n")
+            'abc '
+            >>> remove_whitespace("abc  \\n  ")
+            'abc '
+
+        Use ``leading=True`` to remove leading new line characters, including any surrounding
+        white space:
+
+            >>> remove_whitespace("\\nabc", leading=True)
+            'abc'
+            >>> remove_whitespace("  \\n  abc", leading=True)
+            'abc'
+
+        Use ``trailing=True`` to remove trailing new line characters, including any surrounding
+        white space:
+
+            >>> remove_whitespace("abc  \\n  ", trailing=True)
+            'abc'
+    """
+    # Remove any leading new line characters along with any surrounding white space
+    if leading:
+        string = re.sub(r'^\s*\n+\s*', '', string)
+
+    # Remove any trailing new line characters along with any surrounding white space
+    if trailing:
+        string = re.sub(r'\s*\n+\s*$', '', string)
+
+    # Replace new line characters and absorb any surrounding space.
     string = re.sub(r'\s*\n\s*', ' ', string)
-    return re.sub(r'>\s{2+}<', '><', string)
+    # TODO need some way to get rid of extra spaces in e.g. text <span>   </span>  text
+    return re.sub(r'\s+', ' ', string)
 
 def delete_paragraph(paragraph):
     # https://github.com/python-openxml/python-docx/issues/33#issuecomment-77661907
@@ -68,7 +134,7 @@ def delete_paragraph(paragraph):
     p.getparent().remove(p)
     p._p = p._element = None
 
-fonts = {
+font_styles = {
     'b': 'bold',
     'strong': 'bold',
     'em': 'italic',
@@ -80,6 +146,16 @@ fonts = {
     'th': 'bold',
 }
 
+font_names = {
+    'code': 'Courier',
+    'pre': 'Courier',
+}
+
+styles = {
+    'LIST_BULLET': 'List Bullet',
+    'LIST_NUMBER': 'List Number',
+}
+
 class HtmlToDocx(HTMLParser):
 
     def __init__(self):
@@ -96,6 +172,7 @@ class HtmlToDocx(HTMLParser):
             'table > tbody > tr',
             'table > tfoot > tr'
         ]
+        self.table_style = DEFAULT_TABLE_STYLE
 
     def set_initial_attrs(self, document=None):
         self.tags = {
@@ -116,6 +193,10 @@ class HtmlToDocx(HTMLParser):
         self.skip_tag = None
         self.instances_to_skip = 0
 
+    def copy_settings_from(self, other):
+        """Copy settings from another instance of HtmlToDocx"""
+        self.table_style = other.table_style
+
     def get_cell_html(self, soup):
         # Returns string of td element with opening and closing <td> tags removed
         # Cannot use find_all as it only finds element tags and does not find text which
@@ -180,9 +261,9 @@ class HtmlToDocx(HTMLParser):
             list_type = 'ul' # assign unordered if no tag
 
         if list_type == 'ol':
-            list_style = "List Number"
+            list_style = styles['LIST_NUMBER']
         else:
-            list_style = 'List Bullet'
+            list_style = styles['LIST_BULLET']
 
         self.paragraph = self.doc.add_paragraph(style=list_style)            
         self.paragraph.paragraph_format.left_indent = Inches(min(list_depth * LIST_INDENT, MAX_INDENT))
@@ -237,6 +318,13 @@ class HtmlToDocx(HTMLParser):
         table_soup = self.tables[self.table_no]
         rows, cols = self.get_table_dimensions(table_soup)
         self.table = self.doc.add_table(rows, cols)
+
+        if self.table_style:
+            try:
+                self.table.style = self.table_style
+            except KeyError as e:
+                raise ValueError(f"Unable to apply style {self.table_style}.") from e
+
         rows = self.get_table_rows(table_soup)
         cell_row = 0
         for row in rows:
@@ -248,6 +336,7 @@ class HtmlToDocx(HTMLParser):
                     cell_html = "<b>%s</b>" % cell_html
                 docx_cell = self.table.cell(cell_row, cell_col)
                 child_parser = HtmlToDocx()
+                child_parser.copy_settings_from(self)
                 child_parser.add_html_to_cell(cell_html, docx_cell)
                 cell_col += 1
             cell_row += 1
@@ -275,7 +364,17 @@ class HtmlToDocx(HTMLParser):
         # Create sub-run
         subrun = self.paragraph.add_run()
         rPr = docx.oxml.shared.OxmlElement('w:rPr')
-        rPr.style = 'Hyperlink'
+
+        # add default color
+        c = docx.oxml.shared.OxmlElement('w:color')
+        c.set(docx.oxml.shared.qn('w:val'), "0000EE")
+        rPr.append(c)
+
+        # add underline
+        u = docx.oxml.shared.OxmlElement('w:u')
+        u.set(docx.oxml.shared.qn('w:val'), 'single')
+        rPr.append(u)
+
         subrun._r.append(rPr)
         subrun._r.text = text
 
@@ -309,13 +408,37 @@ class HtmlToDocx(HTMLParser):
             return
 
         self.tags[tag] = current_attrs
-        if tag == 'p':
+        if tag in ['p', 'pre']:
             self.paragraph = self.doc.add_paragraph()
 
         elif tag == 'li':
             self.handle_li()
 
-        elif tag[0] == 'h' and len(tag) == 2:
+        elif tag == "hr":
+
+            # This implementation was taken from:
+            # https://github.com/python-openxml/python-docx/issues/105#issuecomment-62806373
+
+            self.paragraph = self.doc.add_paragraph()
+            pPr = self.paragraph._p.get_or_add_pPr()
+            pBdr = OxmlElement('w:pBdr')
+            pPr.insert_element_before(pBdr,
+                'w:shd', 'w:tabs', 'w:suppressAutoHyphens', 'w:kinsoku', 'w:wordWrap',
+                'w:overflowPunct', 'w:topLinePunct', 'w:autoSpaceDE', 'w:autoSpaceDN',
+                'w:bidi', 'w:adjustRightInd', 'w:snapToGrid', 'w:spacing', 'w:ind',
+                'w:contextualSpacing', 'w:mirrorIndents', 'w:suppressOverlap', 'w:jc',
+                'w:textDirection', 'w:textAlignment', 'w:textboxTightWrap',
+                'w:outlineLvl', 'w:divId', 'w:cnfStyle', 'w:rPr', 'w:sectPr',
+                'w:pPrChange'
+            )
+            bottom = OxmlElement('w:bottom')
+            bottom.set(qn('w:val'), 'single')
+            bottom.set(qn('w:sz'), '6')
+            bottom.set(qn('w:space'), '1')
+            bottom.set(qn('w:color'), 'auto')
+            pBdr.append(bottom)
+
+        elif re.match('h[1-9]', tag):
             if isinstance(self.doc, docx.document.Document):
                 h_size = int(tag[1])
                 self.paragraph = self.doc.add_heading(level=min(h_size, 9))
@@ -331,7 +454,7 @@ class HtmlToDocx(HTMLParser):
             return
 
         # set new run reference point in case of leading line breaks
-        if tag == 'p' or tag == 'li':
+        if tag in ['p', 'li', 'pre']:
             self.run = self.paragraph.add_run()
 
         # add style
@@ -375,6 +498,11 @@ class HtmlToDocx(HTMLParser):
         if self.skip:
             return
 
+        # Only remove white space if we're not in a pre block.
+        if 'pre' not in self.tags:
+            # remove leading and trailing whitespace in all instances
+            data = remove_whitespace(data, True, True)
+
         if not self.paragraph:
             self.paragraph = self.doc.add_paragraph()
 
@@ -393,13 +521,16 @@ class HtmlToDocx(HTMLParser):
                     style = self.parse_dict_string(span['style'])
                     self.add_styles_to_run(style)
 
-
-            # add font style
+            # add font style and name
             for tag in self.tags:
-                if tag in fonts:
-                    font_style = fonts[tag]
+                if tag in font_styles:
+                    font_style = font_styles[tag]
                     setattr(self.run.font, font_style, True)
 
+                if tag in font_names:
+                    font_name = font_names[tag]
+                    self.run.font.name = font_name
+
     def ignore_nested_tables(self, tables_soup):
         """
         Returns array containing only the highest level tables
@@ -446,9 +577,7 @@ class HtmlToDocx(HTMLParser):
     def run_process(self, html):
         if self.bs and BeautifulSoup:
             self.soup = BeautifulSoup(html, 'html.parser')
-            html = remove_whitespace(str(self.soup))
-        else:
-            html = remove_whitespace(html)
+            html = str(self.soup)
         if self.include_tables:
             self.get_tables()
         self.feed(html)
